module("L_Freebox", package.seeall)


local socket = require("socket")
local http = require("socket.http")
local ltn12 = require("ltn12")
local json = require("json")
local sha1 = require("sha1")


local PLUGIN_VERSION = "0.2"

local THIS_DEVICE = 0


local appToken = ""
local sessionToken = ""
local baseUrl = ""
local serverIp = "mafreebox.freebox.fr"
local logged = false
local settingsPerm = false
local refreshRate = 0
local netDevices = nil

local PLUGIN_SID = "urn:freebox-fr:serviceId:Freebox1"

local taskHandle = -1
local TASK_ERROR = 2
local TASK_ERROR_PERM = -2
local TASK_SUCCESS = 4
local TASK_BUSY = 1


local function log(text, level)
	luup.log("Freebox: " .. text, (level or 50))
end


local function task(text, mode)
	if (mode == TASK_ERROR_PERM)
	then
		log(text, 1)
	elseif (mode ~= TASK_SUCCESS)
	then
		log(text, 2)
	else
		log(text)
	end
	if (mode == TASK_ERROR_PERM)
	then
		taskHandle = luup.task(text, TASK_ERROR, "Freebox", taskHandle)
	else
		taskHandle = luup.task(text, mode, "Freebox", taskHandle)

		-- Clear the previous error, since they're all transient
		if (mode ~= TASK_SUCCESS)
		then
			luup.call_delay("clearTask", 15, "", false)
		end
	end
end

function clearTask()
	task("Clearing...", TASK_SUCCESS)
end


local function setVarIfChanged(serviceId, name, value, deviceId, default)
	local curValue = luup.variable_get(serviceId, name, deviceId)
	if ((value ~= curValue) or (curValue == nil)) then
		luup.variable_set(serviceId, name, value, deviceId)
		return true
	else
		return (default or false)
	end
end


local function forceVarChange(serviceId, name, value, deviceId)
	local curValue = luup.variable_get(serviceId, name, deviceId)
	if (value == curValue) then
		luup.variable_set(serviceId, name, "", deviceId)
	end
	luup.variable_set(serviceId, name, value, deviceId)
end


local function setAttrIfChanged(name, value, deviceId, default)
	local curValue = luup.attr_get(name, deviceId)

	if ((value ~= curValue) or (curValue == nil)) then
		luup.attr_set(name, value, deviceId)
		return true
	else
		return (default or false)
	end
end


local sock = function()
	local s = socket.tcp()
	s:settimeout(2)
	return s
end


local function getData(request, code, result)
	local data = {}
	if (request == nil and code ~= "closed") then
		data.success = false
		data.msg = "Freebox Revolution non joignable"
		data.error_code = code
	elseif (code ~= 200 and code ~= 403) then
		data.success = false
		data.msg = "Echec de la communication avec la Freebox Revolution"
		data.error_code = code
	else
		--log(result)
		data = json.decode(result)
		if (data.success == nil) then
			data.success = true
		end
	end
	return data
end


local function getApiBaseUrl()
	local baseUrl = ""
	local url = "http://mafreebox.freebox.fr/api_version"
	local result = {}
	local request, code = http.request({
		url = url,
		create = sock,
		sink = ltn12.sink.table(result)
	})
	local data = getData(request, code, table.concat(result))
	if (data.success) then
		baseUrl = "http://%s" .. data.api_base_url .. "v" .. string.match(data.api_version, "%d+")
		--log("baseUrl " .. baseUrl)
	else
		log("Echec requte: " .. data.msg .. " ( " .. data.error_code .. " )", 2)
	end
	return baseUrl
end


local function restoreAppToken()
	local fh = io.open("/etc/.freebox", "r")
	if (fh == nil) then
		appToken = ""
	else
		appToken = fh:read("*all")
		fh:close()
	end
end

local function saveAppToken(token)
	local fh = io.open("/etc/.freebox", "w+")
	fh:write(token)
	fh:close()
	restoreAppToken()
end


function getAppToken()
	local token = ""
	local statusAuth = "Non authentifie ; appuyer sur le bouton " .. '"' .. "s'authentifier" .. '"'

	if (baseUrl == "") then
		statusAuth = "Freebox Revolution non trouve sur le reseau"
	else
		log("Demande d'authentification ...")
		local url = string.format(baseUrl, serverIp) .. "/login/authorize"
		local requestData = '{ "app_id": "vera.plugin", "app_name": "Freebox Vera Plugin", "app_version": "0.1", "device_name": "Vera" }'
		local result = {}
		local request, code = http.request({
			url = url,
			create = sock,
			method = "POST",
			headers = { ["Content-Type"] = "application/json",
						["Content-Length"] = requestData:len() },
			source = ltn12.source.string(requestData),
			sink = ltn12.sink.table(result)
		})
		local data = getData(request, code, table.concat(result))
		if (data.success) then
			token = data.result.app_token
			local trackId = data.result.track_id
			--log("token " .. token)
			--log("trackId " .. trackId)

			local status
			repeat
				status = "error"
				url = string.format(baseUrl, serverIp) .. "/login/authorize/" .. trackId
				result = {}
				request, code = http.request({
					url = url,
					create = sock,
					sink = ltn12.sink.table(result)
				})
				data = getData(request, code, table.concat(result))
				if (data.success) then
					status = data.result.status
					--log("status " .. status)
					--log("challenge " .. data.result.challenge)
				else
					log("Echec requte: " .. data.msg .. " ( " .. data.error_code .. " )", 2)
				end
				if (status == "pending") then
					log("Authentification en cours ...")
					luup.sleep(2000)
				end
			until status ~= "pending"
			log(status)
			if (status == "granted") then
				log("Authentification ok")
				statusAuth = "Authentifie"
			else
				log("Authentification KO")
				token = ""
			end
		else
			log("Echec requte: " .. data.msg .. " ( " .. data.error_code .. " )", 2)
		end
	end
	saveAppToken(token)
	setVarIfChanged(PLUGIN_SID, "StatusAuthentification", statusAuth, THIS_DEVICE, false)
end


local function openSession()
	logged = false
	local newSessionToken = ""
	local newSettingsPerm = false
	local challenge = nil
	local statusAuth = "Non authentifie ; appuyer sur le bouton " .. '"' .. "s'authentifier" .. '"'
	local statusConfig
	if (settingsPerm) then
		statusConfig = "Autorise"
	else
		statusConfig = "Non autorise"
	end

	if (baseUrl == "") then
		statusAuth = "Freebox Revolution non trouve sur le reseau"
	elseif (appToken ~= "") then
		local url = string.format(baseUrl, serverIp) .. "/login"
		local result = {}
		local headers = {}
		if (sessionToken ~= "") then
			headers["X-Fbx-App-Auth"] = sessionToken
		end
		local request, code = http.request({
			url = url,
			create = sock,
			headers = headers,
			sink = ltn12.sink.table(result)
		})
		local data = getData(request, code, table.concat(result))
		if (data.success) then
			logged = data.result.logged_in
			challenge = data.result.challenge
		else
			log("Echec requte: " .. data.msg .. " ( " .. data.error_code .. " )", 2)
		end

		if (not logged and challenge ~= nil) then
			log("Demande de session ...")
			local password = sha1.hmac_sha1(appToken, challenge)
			url = string.format(baseUrl, serverIp) .. "/login/session"
			local requestData = '{ "app_id": "vera.plugin", "password": "' .. password .. '" }'
			result = {}
			request, code = http.request({
				url = url,
				create = sock,
				method = "POST",
				headers = { ["Content-Type"] = "application/json",
							["Content-Length"] = requestData:len() },
				source = ltn12.source.string(requestData),
				sink = ltn12.sink.table(result)
			})
			data = getData(request, code, table.concat(result))
			if (data.success) then
				log("Session ok")
				newSessionToken = data.result.session_token
				newSettingsPerm = data.result.permissions.settings
				-- log("newSessionToken " .. newSessionToken)
				-- if (newSettingsPerm) then
					-- log("permission ok")
				-- else
					-- log("permission ko")
				-- end
			else
				log("Echec requte: " .. data.msg .. " ( " .. data.error_code .. " )", 2)
				log("Session KO")
				if (data.error_code == "invalid_token") then
					saveAppToken("")
				end
			end

			if (newSessionToken ~= "") then
				url = string.format(baseUrl, serverIp) .. "/login"
				result = {}
				request, code = http.request({
					url = url,
					create = sock,
					headers = { ["X-Fbx-App-Auth"] = newSessionToken },
					sink = ltn12.sink.table(result)
				})
				data = getData(request, code, table.concat(result))
				if (data.success) then
					logged = data.result.logged_in
				else
					log("Echec requte: " .. data.msg .. " ( " .. data.error_code .. " )", 2)
				end
			end
		end
	end

	if (newSessionToken ~= "") then
		sessionToken = newSessionToken
		settingsPerm = newSettingsPerm
	end
	if (logged) then
		statusAuth = "Authentifie et connecte"
		if (settingsPerm) then
			statusConfig = "Autorise"
		else
			statusConfig = "Non autorise"
		end
	end
	setVarIfChanged(PLUGIN_SID, "StatusAuthentification", statusAuth, THIS_DEVICE, false)
	setVarIfChanged(PLUGIN_SID, "StatusConfigFreebox", statusConfig, THIS_DEVICE, false)
end


local function checkLanServer(checkSession)
	if (checkSession) then
		openSession()
		if (not logged) then
			--log("non connect", 2)
			return
		end
	end

	local url = string.format(baseUrl, serverIp) .. "/lan/config"
	local result = {}
	local request, code = http.request({
		url = url,
		create = sock,
		headers = { ["X-Fbx-App-Auth"] = sessionToken },
		sink = ltn12.sink.table(result)
	})
	local data = getData(request, code, table.concat(result))
	if (data.success) then
		if (data.result.name ~= nil and data.result.name ~= "") then
			--log("name " .. data.result.name)
			setAttrIfChanged("name", data.result.name, THIS_DEVICE, false)
		end
		if (data.result.ip ~= nil and data.result.ip ~= "") then
			--log("ip " .. data.result.ip)
			setAttrIfChanged("ip", data.result.ip, THIS_DEVICE, false)
			serverIp = data.result.ip
		end
	else
		log("Echec requte: " .. data.msg .. " ( " .. data.error_code .. " )", 2)
	end
end


local function checkNetworkDevices(checkSession)
	if (checkSession) then
		openSession()
		if (not logged) then
			--log("non connect", 2)
			return
		end
	end

	local url = string.format(baseUrl, serverIp) .. "/lan/browser/interfaces"
	local result = {}
	local request, code = http.request({
		url = url,
		create = sock,
		headers = { ["X-Fbx-App-Auth"] = sessionToken },
		sink = ltn12.sink.table(result)
	})
	local data = getData(request, code, table.concat(result))
	if (data.success) then
		local interfaceName = data.result[1].name
		local count = data.result[1].host_count
		--log("name " .. interfaceName)
		--log("host_count " .. count)
		if (count > 0) then
			url = string.format(baseUrl, serverIp) .. "/lan/browser/" .. interfaceName
			result = {}
			request, code = http.request({
				url = url,
				create = sock,
				headers = { ["X-Fbx-App-Auth"] = sessionToken },
				sink = ltn12.sink.table(result)
			})
			data = getData(request, code, table.concat(result))
			if (data.success) then
				local devices = {}
				for i=1,count do
					local id = data.result[i].id
					devices[id] = {}
					devices[id].name = data.result[i].primary_name
					devices[id].mac = ""
					if (data.result[i].l2ident ~= nil
							and data.result[i].l2ident.type == "mac_address") then
						devices[id].mac = data.result[i].l2ident.id
					end
					devices[id].ip = ""
					if (data.result[i].l3connectivities ~= nil) then
						for j=1,5 do
							if (data.result[i].l3connectivities[j] ~= nil
									and data.result[i].l3connectivities[j].af == "ipv4") then
								devices[id].ip = data.result[i].l3connectivities[j].addr
								break
							end
						end
					end
					devices[id].reachable = data.result[i].reachable
					if (netDevices ~= nil and netDevices[id] ~= nil and devices[id].mac ~= "") then
						if (netDevices[id].reachable and not devices[id].reachable) then
							forceVarChange(PLUGIN_SID, "PeripheriqueInjoignable", devices[id].mac, THIS_DEVICE)
						elseif (not netDevices[id].reachable and devices[id].reachable) then
							forceVarChange(PLUGIN_SID, "PeripheriqueJoignable", devices[id].mac, THIS_DEVICE)
						end
					end
				end
				netDevices = devices
			else
				log("Echec requte: " .. data.msg .. " ( " .. data.error_code .. " )", 2)
			end
		end
	else
		log("Echec requte: " .. data.msg .. " ( " .. data.error_code .. " )", 2)
	end

	local xml = "<net_devices>"
	if (netDevices ~= nil) then
		for k, v in pairs(netDevices) do
			if (v.reachable) then
				xml = xml .. "<device>"
				xml = xml .. "<name>" .. v.name .. "</name>"
				xml = xml .. "<ip>" .. v.ip .. "</ip>"
				xml = xml .. "<mac>" .. v.mac .. "</mac>"
				xml = xml .. "<reachable>true</reachable>"
				xml = xml .. "</device>"
			end
		end
		for k, v in pairs(netDevices) do
			if (not v.reachable) then
				xml = xml .. "<device>"
				xml = xml .. "<name>" .. v.name .. "</name>"
				xml = xml .. "<ip>" .. v.ip .. "</ip>"
				xml = xml .. "<mac>" .. v.mac .. "</mac>"
				xml = xml .. "<reachable>false</reachable>"
				xml = xml .. "</device>"
			end
		end
	end
	xml = xml .. "</net_devices>"
	setVarIfChanged(PLUGIN_SID, "PeripheriquesReseau", xml, THIS_DEVICE, false)
end


function wifi(checkSession, requestData)
	if (checkSession) then
		openSession()
		if (not logged) then
			--log("non connect", 2)
			return
		end
	end

	local url = string.format(baseUrl, serverIp) .. "/wifi/config"
	local request, code
	local result = {}
	if (requestData ~= nil and requestData ~= "") then
		request, code = http.request({
			url = url,
			create = sock,
			method = "PUT",
			headers = { ["X-Fbx-App-Auth"] = sessionToken,
						["Content-Type"] = "application/json",
						["Content-Length"] = requestData:len() },
			source = ltn12.source.string(requestData),
			sink = ltn12.sink.table(result)
		})
	else
		request, code = http.request({
			url = url,
			create = sock,
			headers = { ["X-Fbx-App-Auth"] = sessionToken },
			sink = ltn12.sink.table(result)
		})
	end
	local data = getData(request, code, table.concat(result))
	if (data.success) then
		if (data.result.ap_params.enabled) then
			--log("Wifi actif")
			setVarIfChanged(PLUGIN_SID, "WifiActif", "1", THIS_DEVICE, false)
		else
			--log("Wifi inactif")
			setVarIfChanged(PLUGIN_SID, "WifiActif", "0", THIS_DEVICE, false)
		end
	else
		log("Echec requte: " .. data.msg .. " ( " .. data.error_code .. " )", 2)
		task(data.msg, TASK_ERROR)
	end
end


local function checkWifi(checkSession)
	wifi(checkSession, "")
end


function enableWifi()
	wifi(true, '{ "ap_params": { "enabled": true } }')
end


function disableWifi()
	wifi(true, '{ "ap_params": { "enabled": false } }')
end


function ftp(checkSession, requestData)
	if (checkSession) then
		openSession()
		if (not logged) then
			--log("non connect", 2)
			return
		end
	end

	local url = string.format(baseUrl, serverIp) .. "/ftp/config"
	local request, code
	local result = {}
	if (requestData ~= nil and requestData ~= "") then
		request, code = http.request({
			url = url,
			create = sock,
			method = "PUT",
			headers = { ["X-Fbx-App-Auth"] = sessionToken,
						["Content-Type"] = "application/json",
						["Content-Length"] = requestData:len() },
			source = ltn12.source.string(requestData),
			sink = ltn12.sink.table(result)
		})
	else
		request, code = http.request({
			url = url,
			create = sock,
			headers = { ["X-Fbx-App-Auth"] = sessionToken },
			sink = ltn12.sink.table(result)
		})
	end
	local data = getData(request, code, table.concat(result))
	if (data.success) then
		if (data.result.enabled) then
			--log("FTP actif")
			setVarIfChanged(PLUGIN_SID, "FtpActif", "1", THIS_DEVICE, false)
		else
			--log("FTP inactif")
			setVarIfChanged(PLUGIN_SID, "FtpActif", "0", THIS_DEVICE, false)
		end
	else
		log("Echec requte: " .. data.msg .. " ( " .. data.error_code .. " )", 2)
		task(data.msg, TASK_ERROR)
	end
end


local function checkFtp(checkSession)
	ftp(checkSession, "")
end


function enableFtp()
	ftp(true, '{ "enabled": true }')
end


function disableFtp()
	ftp(true, '{ "enabled": false }')
end


function sambaShares(checkSession, requestData)
	if (checkSession) then
		openSession()
		if (not logged) then
			--log("non connect", 2)
			return
		end
	end

	local url = string.format(baseUrl, serverIp) .. "/netshare/samba"
	local request, code
	local result = {}
	if (requestData ~= nil and requestData ~= "") then
		request, code = http.request({
			url = url,
			create = sock,
			method = "PUT",
			headers = { ["X-Fbx-App-Auth"] = sessionToken,
						["Content-Type"] = "application/json",
						["Content-Length"] = requestData:len() },
			source = ltn12.source.string(requestData),
			sink = ltn12.sink.table(result)
		})
	else
		request, code = http.request({
			url = url,
			create = sock,
			headers = { ["X-Fbx-App-Auth"] = sessionToken },
			sink = ltn12.sink.table(result)
		})
	end
	local data = getData(request, code, table.concat(result))
	if (data.success) then
		if (data.result.file_share_enabled) then
			--log("Partage fichiers Windows actif")
			setVarIfChanged(PLUGIN_SID, "PartageFichiersWinActif", "1", THIS_DEVICE, false)
		else
			--log("Partage fichiers Windows inactif")
			setVarIfChanged(PLUGIN_SID, "PartageFichiersWinActif", "0", THIS_DEVICE, false)
		end
		if (data.result.print_share_enabled) then
			--log("Partage imprimantes Windows actif")
			setVarIfChanged(PLUGIN_SID, "PartageImprimantesWinActif", "1", THIS_DEVICE, false)
		else
			--log("Partage imprimantes Windows inactif")
			setVarIfChanged(PLUGIN_SID, "PartageImprimantesWinActif", "0", THIS_DEVICE, false)
		end
	else
		log("Echec requte: " .. data.msg .. " ( " .. data.error_code .. " )", 2)
		task(data.msg, TASK_ERROR)
	end
end


local function checkSambaShares(checkSession)
	sambaShares(checkSession, "")
end


function enableSambaFiles()
	sambaShares(true, '{ "file_share_enabled": true }')
end


function disableSambaFiles()
	sambaShares(true, '{ "file_share_enabled": false }')
end


function enableSambaPrinters()
	sambaShares(true, '{ "print_share_enabled": true }')
end


function disableSambaPrinters()
	sambaShares(true, '{ "print_share_enabled": false }')
end


local function getRefreshRate()
	local rate = tonumber(luup.variable_get(PLUGIN_SID, "FrequenceMiseAJour", THIS_DEVICE) or "")
	if (rate == nil or (rate > 0 and rate < 5)) then
		rate = 0
		setVarIfChanged(PLUGIN_SID, "FrequenceMiseAJour", string.format("%d", rate), THIS_DEVICE, false)
	end
	refreshRate = rate
end


function setRefreshRate(rate)
	rate = tonumber(rate)
	if (rate ~= nil) then
		if (rate < 5) then
			rate = 0
		end
		setVarIfChanged(PLUGIN_SID, "FrequenceMiseAJour", string.format("%d", rate), THIS_DEVICE, false)
		local oldRate = refreshRate
		refreshRate = rate
		if (oldRate < 5 and refreshRate >= 5) then
			refreshData()
		end
	end
end


-- Function called at plugin startup
function startup(lul_device)

	THIS_DEVICE = lul_device

	task("Dmarrage plugin Freebox", TASK_SUCCESS)
	
	luup.call_delay("deferredFreeboxStartup", 1)

	task("Plugin Freebox ok", TASK_SUCCESS)
	
	return true

end


function deferredStartup(data)

	setVarIfChanged(PLUGIN_SID, "PluginVersion", PLUGIN_VERSION, THIS_DEVICE, false)

	baseUrl = getApiBaseUrl()

	restoreAppToken()

	getRefreshRate()

	refreshData()

end


function refreshData()
	--log("mise  jour des donnes")

	if (refreshRate >= 5) then
		luup.call_delay("refreshData", refreshRate)
	end

	openSession()
	if (not logged) then
		--log("non connect", 2)
		return
	end

	checkLanServer(false)

	checkNetworkDevices(false)

	-- local devices = luup.variable_get("urn:freebox-fr:serviceId:Freebox1", "PeripheriquesReseau", THIS_DEVICE) or ""
	-- local mac = ""
	-- local reachable = devices:match("<mac>"..mac.."</mac><reachable>(.-)</reachable>")
	-- if (reachable == "true") then
		-- log(mac .. " joignable")
	-- elseif (reachable == "false") then
		-- log(mac .. " injoignable")
	-- else
		-- log(mac .. " inconnu")
	-- end
	-- local ip = ""
	-- reachable = devices:match("<ip>"..ip.."</ip>.-<reachable>(.-)</reachable>")
	-- if (reachable == "true") then
		-- log(ip .. " joignable")
	-- elseif (reachable == "false") then
		-- log(ip .. " injoignable")
	-- else
		-- log(ip .. " inconnu")
	-- end

	checkWifi(false)

	checkFtp(false)

	checkSambaShares(false)
end
